'''
Create clusters from an input file
'''

from wkmeans_geo.src import clusters as cl
import pandas as pd


def create_initial_clusters_from_given_input(number_of_clusters, initial_solution):
    '''
    Function to create initial clusters from input
    :param number_of_clusters:
    :param initial_solution:
    :return:
    '''

    given_number_of_clusters = len(initial_solution['CLUSTER'].unique())

    if number_of_clusters > given_number_of_clusters:

        initial_solution_clusters = cl.calculate_cluster_centers(initial_solution)
        initial_solution_with_clusters = initial_solution.merge(initial_solution_clusters, how='left', on='CLUSTER')
        initial_solution_with_clusters = cl.calculate_weighted_distance(initial_solution_with_clusters )

        initial_solution_summary = initial_solution_with_clusters.groupby(['CLUSTER'], as_index=False).agg({'LOCATION_NAME': 'count', 'WEIGHTED_DISTANCE': 'sum'})
        initial_solution_summary.rename(columns={'LOCATION_NAME': 'NUMBER_OF_LOCATIONS', 'WEIGHTED_DISTANCE': 'TOTAL_WEIGHTED_DISTANCE'}, inplace=True)
        initial_solution_summary.sort_values(by=['TOTAL_WEIGHTED_DISTANCE'], ascending=False, inplace=True)

        # filter clusters where number of locations is twice as big as given cluster and number of cluster difference
        filter_clusters = initial_solution_summary['NUMBER_OF_LOCATIONS'] > 2*(number_of_clusters - given_number_of_clusters)
        top_clusters = initial_solution_summary[filter_clusters]
        top_clusters = top_clusters.head(1)

        #
        top_cluster = initial_solution_with_clusters[initial_solution_with_clusters['CLUSTER']==top_clusters['CLUSTER'].iloc[0]]
        top_cluster.sort_values(by=['WEIGHTED_DISTANCE'], ascending=False, inplace=True)

        clusters_added = range(1, (number_of_clusters - given_number_of_clusters)+1)
        idx=0
        cluster_no = given_number_of_clusters
        new_location_store = []
        for cluster in clusters_added:
            cluster_no = cluster_no + 1
            new_location_store.append(pd.DataFrame([{'LOCATION_NAME': top_cluster.iloc[idx]['LOCATION_NAME'],
                                        'NEW_CLUSTER': str(cluster_no)}]))
            idx=idx+1
        new_location_store = pd.concat(new_location_store)

        initial_solution = initial_solution.merge(new_location_store, how='left')
        filter_new_clusters = initial_solution['NEW_CLUSTER'].notnull()
        initial_solution.loc[filter_new_clusters, 'CLUSTER'] = initial_solution.loc[filter_new_clusters, 'NEW_CLUSTER']
        initial_solution.drop(['NEW_CLUSTER'], 1, inplace=True)

        return initial_solution

    else:

        raise Exception('Number of Clusters should be more than total number of clusters in the initial solution')


'''
Create test data for testing
'''

import pandas as pd
import random
from faker import Faker
fake = Faker()

number_of_locations = 100
number_of_clusters = 3
minimum_elements_in_a_cluster = 0
maximum_elements_in_a_cluster = 200
maximum_volume_in_a_cluster = 100
maximum_iteration = 100
objective_range = 0.001
enable_minimum_maximum_elements_in_a_cluster = True

# create locations test
locations_with_clusters_test = []
for location in range(0, number_of_locations):
    locations_with_clusters_test.append({'LOCATION_NAME': 'LOC {}'.format(str(location)),
                                         'LATITUDE': float(fake.latitude()),
                                         'LONGITUDE': float(fake.longitude()),
                                         'VOLUME': 1,
                                         'WEIGHT': 1,
                                         'CLUSTER': random.randint(1, number_of_clusters)})
locations_with_clusters_test = pd.DataFrame(locations_with_clusters_test)
locations_test = locations_with_clusters_test.drop(['CLUSTER'], 1)

del location
del number_of_locations
del fake



import os
import sys
#sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '/src')))
#sys.path.append('src')

import pandas as pd
from wkmeans_geo.src import clusters as cl
from wkmeans_geo.src import initiation as init
from wkmeans_geo.src import optimization_model as ort


def calculate_clusters(input_locations,
                       number_of_clusters,
                       minimum_elements_in_a_cluster,
                       maximum_elements_in_a_cluster,
                       maximum_volume_in_a_cluster,
                       maximum_iteration,
                       objective_range,
                       enable_minimum_maximum_elements_in_a_cluster):
    '''
    Run WKmens clustering
    :param input_locations:
    :param number_of_clusters:
    :param minimum_elements_in_a_cluster:
    :param maximum_elements_in_a_cluster:
    :param maximum_volume_in_a_cluster:
    :param maximum_iteration:
    :param objective_range:
    :param enable_minimum_maximum_elements_in_a_cluster:
    :param previous_objective:
    :return:
    '''

    iteration = 0
    all_clusters = []
    all_stores_with_clusters = []

    if 'CLUSTER' not in input_locations.columns:
        locations_with_clusters = init.randomly_assign_clusters(number_of_clusters, input_locations)
    else:
        locations_with_clusters = input_locations.copy()
        input_locations.drop(['CLUSTER'], 1, inplace=True)

    prev_objective, prev_clusters, prev_locations_with_clusters = \
        init.initiate_algorithm(iteration, locations_with_clusters)

    all_clusters.append(prev_clusters)
    all_stores_with_clusters.append(prev_locations_with_clusters)

    solution_not_found = True
    while solution_not_found:
        #'Running iteration {}'.format(str(iteration)

        location_cluster_distance_matrix = cl.calculate_distance_matrix(input_locations.copy(), prev_clusters)

        location_list, cluster_list, distance, volume = ort.prepare_model_inputs(location_cluster_distance_matrix)
        solution = ort.formulate_and_solve_ortools_model(location_list,
                                                         cluster_list,
                                                         distance,
                                                         volume,
                                                         minimum_elements_in_a_cluster,
                                                         maximum_elements_in_a_cluster,
                                                         maximum_volume_in_a_cluster,
                                                         enable_minimum_maximum_elements_in_a_cluster)

        if len(solution) > 0:

            #Optimal solution is found'

            iteration = iteration + 1
            solution['ITERATION'] = iteration
            objective = round(solution['WEIGHTED_DISTANCE'].sum(), 2)
            solution['OBJECTIVE'] = objective

            locations_with_clusters = input_locations.copy().merge(solution, how='left', on=['LOCATION_NAME'])

            #Clusters
            cluster_locations = cl.calculate_cluster_centers(locations_with_clusters)
            cluster_locations['ITERATION'] = iteration

            #Merging results with clusters
            locations_with_clusters = locations_with_clusters.merge(cluster_locations, how='left', on=['CLUSTER'])
            locations_with_clusters['ITERATION'] = iteration
            locations_with_clusters['SOLUTION'] = 0

            locations_with_clusters = locations_with_clusters[
                ['LOCATION_NAME', 'LATITUDE', 'LONGITUDE', 'WEIGHT', 'VOLUME',
                 'CLUSTER', 'CLUSTER_LATITUDE', 'CLUSTER_LONGITUDE',
                 'DISTANCE', 'WEIGHTED_DISTANCE', 'ITERATION',
                 'OBJECTIVE', 'SOLUTION']]

            #'Current objective {}'.format(str(objective))
            #'Prev objective {}'.format(str(prev_objective))

            if abs(objective - prev_objective) < objective_range:
                #'Solution found'
                solution_not_found = False
                prev_locations_with_clusters['SOLUTION'] = 1

            elif (prev_objective < objective) and iteration > maximum_iteration:
                #'Stopping'
                solution_not_found = False
                prev_locations_with_clusters['SOLUTION'] = 1

            else:
                prev_objective = objective
                prev_clusters = cluster_locations
                prev_locations_with_clusters = locations_with_clusters

                all_clusters.append(prev_clusters)
                all_stores_with_clusters.append(prev_locations_with_clusters)

        else:

            raise Exception('Optimal solution to the allocation model does not exist')

    all_clusters = pd.concat(all_clusters)
    all_clusters['NUMBER_OF_CLUSTERS'] = number_of_clusters
    all_stores_with_clusters = pd.concat(all_stores_with_clusters)
    all_stores_with_clusters['NUMBER_OF_CLUSTERS'] = number_of_clusters

    return all_clusters, all_stores_with_clusters




from helpers import haversine_distance


def calculate_cluster_centers(locations_with_clusters):
    '''
    Calculate cluster centers
    :param locations_with_clusters:
    :return:
    '''

    cluster_lat_lons = locations_with_clusters.groupby(['CLUSTER'], as_index=False).agg(
        {'LATITUDE': 'sum', 'LONGITUDE': 'sum',
         'LOCATION_NAME': 'count'}).rename(
        columns={'LATITUDE': 'NEW_CLUSTER_LATITUDE', 'LONGITUDE': 'NEW_CLUSTER_LONGITUDE',
                 'LOCATION_NAME': 'NUMBER_OF_CUSTOMERS_IN_A_CLUSTER'})

    cluster_lat_lons['NEW_CLUSTER_LATITUDE'] = cluster_lat_lons['NEW_CLUSTER_LATITUDE'] / cluster_lat_lons['NUMBER_OF_CUSTOMERS_IN_A_CLUSTER']
    cluster_lat_lons['NEW_CLUSTER_LONGITUDE'] = cluster_lat_lons['NEW_CLUSTER_LONGITUDE'] / cluster_lat_lons['NUMBER_OF_CUSTOMERS_IN_A_CLUSTER']

    cluster_lat_lons['CLUSTER_LATITUDE'] = cluster_lat_lons['NEW_CLUSTER_LATITUDE']
    cluster_lat_lons['CLUSTER_LONGITUDE'] = cluster_lat_lons['NEW_CLUSTER_LONGITUDE']
    cluster_lat_lons.drop(['NEW_CLUSTER_LATITUDE', 'NEW_CLUSTER_LONGITUDE', 'NUMBER_OF_CUSTOMERS_IN_A_CLUSTER'], 1, inplace=True)

    return cluster_lat_lons


def calculate_weighted_distance(locations_with_clusters):
    '''
    Calculate weighted distance
    :param locations_with_clusters:
    :return:
    '''

    distance = haversine_distance.calculate_haversine_distance(locations_with_clusters['LATITUDE'],
                                                               locations_with_clusters['LONGITUDE'],
                                                               locations_with_clusters['CLUSTER_LATITUDE'],
                                                               locations_with_clusters['CLUSTER_LONGITUDE'])
    locations_with_clusters['DISTANCE'] = distance
    locations_with_clusters['WEIGHTED_DISTANCE'] = locations_with_clusters['DISTANCE'] * locations_with_clusters['WEIGHT']

    return locations_with_clusters


def calculate_distance_matrix(locations_at_iteration, clusters_for_distance):
    '''
    Calculate distance matrix
    :param locations_at_iteration:
    :param clusters_for_distance:
    :return:
    '''

    locations_at_iteration['Key'] = 0
    clusters_for_distance['Key'] = 0
    locations_at_iteration_with_clusters = locations_at_iteration.merge(clusters_for_distance)
    locations_at_iteration_with_clusters = calculate_weighted_distance(locations_at_iteration_with_clusters)

    return locations_at_iteration_with_clusters [['LOCATION_NAME', 'CLUSTER', 'DISTANCE', 'WEIGHTED_DISTANCE', 'VOLUME']]



'''
Test class for testing WKmens
'''

import unittest
import os
import sys

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../wkmeans_geo')))
from wkmeans_geo import wkmeans_clustering as wkc
from wkmeans_geo.src import data as test_data
from wkmeans_geo import clusters_from_input as cf


class WKMeansTest(unittest.TestCase):

    def set_up(self):
        self.input_locations = test_data.locations_test
        self.input_locations_with_clusters = test_data.locations_with_clusters_test
        self.number_of_clusters = len(test_data.locations_with_clusters_test.CLUSTER.unique())
        self.minimum_elements_in_a_cluster = test_data.minimum_elements_in_a_cluster
        self.maximum_elements_in_a_cluster = test_data.maximum_elements_in_a_cluster
        self.maximum_volume_in_a_cluster = test_data.maximum_volume_in_a_cluster
        self.maximum_iteration = test_data.maximum_iteration
        self.objective_range = test_data.objective_range
        self.enable_minimum_maximum_elements_in_a_cluster = test_data.enable_minimum_maximum_elements_in_a_cluster

    def test_wkmeans_clustering(self):
        '''
        Test for WKmeans
        :return:
        '''

        self.set_up()

        all_clusters, all_locations_with_clusters = wkc.calculate_clusters(self.input_locations,
                                                                           self.number_of_clusters,
                                                                           self.minimum_elements_in_a_cluster,
                                                                           self.maximum_elements_in_a_cluster,
                                                                           self.maximum_volume_in_a_cluster,
                                                                           self.maximum_iteration,
                                                                           self.enable_minimum_maximum_elements_in_a_cluster,
                                                                           self.objective_range)

        self.assertTrue(len(all_locations_with_clusters) > 0)

    def test_create_initial_clusters_from_given_input(self):
        '''
        To test creating initial clusters from given solution input
        :return:
        '''
        self.set_up()

        initial_solution = cf.create_initial_clusters_from_given_input(self.number_of_clusters + 1,
                                                                       self.input_locations_with_clusters)

        self.assertTrue(len(initial_solution) > 0)

    def test_wkmeans_clustering_with_initial_solution(self):
        '''
        Test to create wkmeans clustering with given initial solution
        :return:
        '''

        self.set_up()

        all_clusters, all_locations_with_clusters = wkc.calculate_clusters(self.input_locations_with_clusters,
                                                                           self.number_of_clusters + 1,
                                                                           self.minimum_elements_in_a_cluster,
                                                                           self.maximum_elements_in_a_cluster,
                                                                           self.maximum_volume_in_a_cluster,
                                                                           self.maximum_iteration,
                                                                           self.enable_minimum_maximum_elements_in_a_cluster,
                                                                           self.objective_range)

        self.assertTrue(len(all_locations_with_clusters) > 0)


if __name__ == '__main__':
    unittest.main()



'''
Formulate and solve depot-store allocation model
'''

import pandas as pd
import collections
from ortools.linear_solver import pywraplp


def prepare_model_inputs(location_cluster_distance_matrix):
    '''
    Run the optimization model
    :param location_cluster_distance_matrix:
    :return:
    '''

    DistanceInputs = collections.namedtuple('DistanceInputs', ['DISTANCE', 'WEIGHTED_DISTANCE'])
    distance = {}
    for row in location_cluster_distance_matrix.itertuples():
        distance[row.CLUSTER, row.LOCATION_NAME] = (DistanceInputs(
            DISTANCE=row.DISTANCE,
            WEIGHTED_DISTANCE=row.WEIGHTED_DISTANCE))

    location_list = []
    cluster_list = []
    for cluster, store in distance.keys():
        cluster_list.append(cluster)
        location_list.append(store)
    location_list = list(dict.fromkeys(location_list))
    cluster_list = list(dict.fromkeys(cluster_list))

    volume_input = location_cluster_distance_matrix[['LOCATION_NAME', 'VOLUME']].drop_duplicates()
    volume = {}
    for _, row in volume_input.iterrows():
        volume[row.LOCATION_NAME] = row.VOLUME

    return location_list, cluster_list, distance, volume


def formulate_and_solve_ortools_model(store_list,
                                      cluster_list,
                                      distance,
                                      volume,
                                      minimum_elements_in_a_cluster,
                                      maximum_elements_in_a_cluster,
                                      maximum_volume_in_a_cluster,
                                      enable_minimum_maximum_elements_in_a_cluster):
    '''
    Formulate the model
    :param store_list:
    :param cluster_list:
    :param distance:
    :param volume:
    :param minimum_elements_in_a_cluster:
    :param maximum_elements_in_a_cluster:
    :param maximum_volume_in_a_cluster:
    :param enable_minimum_maximum_elements_in_a_cluster:
    :return:
    '''

    # formulate model
    solver = pywraplp.Solver('SolveIntegerProblem', pywraplp.Solver.CBC_MIXED_INTEGER_PROGRAMMING)

    # create variables #
    y = {}
    for cluster, store in distance.keys():
        y[cluster, store] = solver.BoolVar('y[cluster = {}, {}]'.format(str(cluster), store))

    # Add constraints
    # each store is assigned to one cluster
    for store in store_list:
        solver.Add(solver.Sum([y[cluster, store] for cluster in cluster_list]) == 1)

    if enable_minimum_maximum_elements_in_a_cluster:
        # minimum number of elements in a cluster
        for cluster in cluster_list:
            solver.Add(solver.Sum([y[cluster, store] for store in store_list]) >= minimum_elements_in_a_cluster)

        # maximum number of elements in a cluster
        for cluster in cluster_list:
            solver.Add(solver.Sum([y[cluster, store] for store in store_list]) <= maximum_elements_in_a_cluster)
    else:

        # minimum number of elements in a cluster
        for cluster in cluster_list:
            solver.Add(solver.Sum([y[cluster, store] for store in store_list]) >= 1)

    if maximum_volume_in_a_cluster is not None:
        for cluster in cluster_list:
            solver.Add(
                solver.Sum([volume[store] * y[cluster, store] for store in store_list]) <= maximum_volume_in_a_cluster)

    # add objective
    solver.Minimize(solver.Sum(
        [distance[cluster, store].WEIGHTED_DISTANCE * y[cluster, store] for cluster, store in
         distance.keys()]))

    # solver.Minimize(1)
    solver_parameters = pywraplp.MPSolverParameters()
    solver_parameters.SetDoubleParam(pywraplp.MPSolverParameters.RELATIVE_MIP_GAP, 0.01)
    # solver.SetTimeLimit(self.solver_time_limit)

    solver.EnableOutput()

    solution = solver.Solve()

    # get solution
    if solution == pywraplp.Solver.OPTIMAL:

        # 'Problem solved in {} milliseconds'.format(str(solver.WallTime())))
        # Problem solved in {} iterations'.format(str(solver.Iterations())))

        solution_final = []
        for cluster, store in distance.keys():
            solution_final.append({'CLUSTER': cluster,
                                   'LOCATION_NAME': store,
                                   'VALUE': y[cluster, store].solution_value(),
                                   'DISTANCE': distance[cluster, store].DISTANCE,
                                   'WEIGHTED_DISTANCE': distance[cluster, store].WEIGHTED_DISTANCE,
                                   'VOLUME': volume[store]})

        solution = pd.DataFrame(solution_final)

        solution = solution[solution['VALUE'] == 1]
        solution = solution[['CLUSTER', 'LOCATION_NAME', 'DISTANCE', 'WEIGHTED_DISTANCE']]

    else:
        solution = pd.DataFrame()

    return solution



from wkmeans_geo.src import clusters as cl


def randomly_assign_clusters(number_of_clusters, input_locations):
    '''
    Randomly assign locations to clusters
    :param number_of_clusters:
    :param input_locations:
    :return:
    '''
    # randomly assign customers to K clusters
    clusters = list(range(0, number_of_clusters)) * round(
        len(input_locations) / number_of_clusters + 1)

    clusters = clusters[:len(input_locations)]
    locations_with_clusters = input_locations.copy()
    locations_with_clusters['CLUSTER'] = clusters
    locations_with_clusters['CLUSTER'] = locations_with_clusters['CLUSTER'].astype(str)

    return locations_with_clusters


def initiate_algorithm(iteration,
                       locations_with_clusters):
    '''
    Initiatie the algorithm
    :param iteration:
    :param locations_with_clusters:
    :return:
    '''

    clusters = cl.calculate_cluster_centers(locations_with_clusters)

    locations_with_clusters = locations_with_clusters.merge(clusters, how='left', on='CLUSTER')

    locations_with_clusters = cl.calculate_weighted_distance(locations_with_clusters)
    locations_with_clusters['ITERATION'] = iteration

    objective = locations_with_clusters['WEIGHTED_DISTANCE'].sum().round(2)
    locations_with_clusters['OBJECTIVE'] = objective

    clusters['ITERATION'] = iteration
    locations_with_clusters['SOLUTION'] = 0

    return objective, clusters, locations_with_clusters



import pathlib
from setuptools import setup, find_packages


def parse_requirements(filename):
    '''
    Function to parse requirements.txt
    :param filename:
    :return:
    '''

    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]


install_reqs = parse_requirements('requirements.txt')
reqs = install_reqs

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="cimren-wkmeans-geo",
    version="1.3.2",
    description="Weighted KMeans Clustering for Geolocational Problem",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/emrahcimren/wkmeans-geo",
    author="cimren",
    author_email="cimren.1@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
    packages=["wkmeans_geo", "wkmeans_geo/src"],
    include_package_data=True,
    install_requires=reqs,
    #entry_points={
    #    "console_scripts": [
    #        "realpython=reader.__main__:main",
    #    ]
    #},
)
